/*

NeatUpload - an HttpModule and User Controls for uploading large files
Copyright (C) 2005,2006  Dean Brettle

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

using System;
using System.Collections;
using System.Collections.Specialized;
using System.IO;
using System.Web;
using System.Web.UI;
using System.Web.UI.Design;
using System.ComponentModel;
using System.Web.UI.WebControls;
using System.Web.UI.HtmlControls;
using System.Security.Permissions;
using System.Text.RegularExpressions;
using Brettle.Web.NeatUpload.Internal.UI;

namespace Brettle.Web.NeatUpload
{	
	/// <summary>
	/// Hidden form field which tells the <see cref="UploadHttpModule"/> to stream all files in the form
	/// to storage and allows that <see cref="ProgressBar"/> control to track progress.
	/// </summary>
	/// <remarks>
	/// You only need to use this control if you are using something other than NeatUpload's controls to
	/// generate your file upload fields.  For example, if you are using client-side script to dynamically
	/// add &lt;input type="file"%gt; elements, you would need to use this control.  If you use this control, you
	/// need to place it before the &lt;input type="file"%gt; elements, preferably immediately after the form
	/// element.
	/// </remarks>
	[AspNetHostingPermissionAttribute (SecurityAction.LinkDemand, Level = AspNetHostingPermissionLevel.Minimal)]
	[AspNetHostingPermissionAttribute (SecurityAction.InheritanceDemand, Level = AspNetHostingPermissionLevel.Minimal)]
	public class HiddenPostBackID : System.Web.UI.WebControls.WebControl
	{

		// Create a logger for use in this class
		/*
		private static readonly log4net.ILog log
			= log4net.LogManager.GetLogger(System.Reflection.MethodBase.GetCurrentMethod().DeclaringType);
		*/
		private bool IsDesignTime = (HttpContext.Current == null);
		
		protected override void OnInit(EventArgs e)
		{
			InitializeComponent();
			base.OnInit(e);
		}
		
		private void InitializeComponent()
		{
			this.Load += new System.EventHandler(this.Control_Load);
		}
				
		private void Control_Load(object sender, EventArgs e)
		{
			if (IsDesignTime)
				return;
			
			// If we can find the containing HtmlForm control, set enctype="multipart/form-data" method="Post".
			// If we can't find it, the page might be using some other form control or not using runat="server",
			// so we assume the developer has already set the enctype and method attributes correctly.
			Control c = Parent;
			while (c != null && !(c is HtmlForm))
			{
				c = c.Parent;
			}
			HtmlForm form = c as HtmlForm;
			if (form != null)
			{
				form.Enctype = "multipart/form-data";
				form.Method = "Post";
			}
		}
		
		protected override void OnPreRender (EventArgs e)
		{
			if (!IsDesignTime && UploadModule.IsEnabled)
			{
				if (!Page.IsClientScriptBlockRegistered("NeatUploadJs"))
				{
					Page.RegisterClientScriptBlock("NeatUploadJs", @"
	<script type='text/javascript' language='javascript' src='" 
                        + UploadModule.GetCacheBustedPath("/NeatUpload/NeatUpload.js")
                        + @"'></script>");
				}
			}
			base.OnPreRender(e);
		}
		
		protected override void Render(HtmlTextWriter writer)
		{
			if (!IsDesignTime && UploadModule.IsEnabled)
			{

				this.Page.RegisterStartupScript("NeatUploadHiddenPostBackID-" + this.UniqueID, @"
<script type='text/javascript' language='javascript'>
<!--
NeatUploadHiddenPostBackIDCreate('" + this.ClientID + @"','"
							 + FormContext.Current.PostBackID + @"');
// -->
</script>");
			}
			writer.AddAttribute(HtmlTextWriterAttribute.Id, ClientID);
			writer.AddAttribute(HtmlTextWriterAttribute.Type, "hidden");
			writer.AddAttribute(HtmlTextWriterAttribute.Name, UploadModule.PostBackIDFieldName);
			if (!IsDesignTime)
			{
				writer.AddAttribute(HtmlTextWriterAttribute.Value, FormContext.Current.PostBackID);
			}
			writer.RenderBeginTag(HtmlTextWriterTag.Input);
			writer.RenderEndTag();
		}
		
	}
}
